""" Functions for evaluating segmentations.
"""
import re
import sys
import numpy as np
import itertools
from collections import defaultdict, Counter
from tabulate import tabulate

encoding    = "utf-8"
verbose = False
comment_pattern = re.compile(ur'[;#].*$', re.MULTILINE)


def print_error(err_text):
    print >>sys.stderr, err_text

def print_if_verbose(text, verbose):
    if verbose:
        print unicode(text).encode(encoding)


def unicodize(lst):
    return [[unicode(el) for el in row] for row in lst]
            
TABLE_FMT = 'orgtbl'
FLOAT_FMT = '.2f'

def utabulate(lst, headers=None, tablefmt=TABLE_FMT):
    if headers:
        return tabulate(unicodize(lst), headers=headers, tablefmt=tablefmt, floatfmt=FLOAT_FMT)
    else:
        return tabulate(unicodize(lst), tablefmt=tablefmt, floatfmt=FLOAT_FMT)

def read_label_file(labels_file):
    """Read label file and allowed choices.
    """
    lines = []
    norm_pattern = re.compile(ur'\s+', re.MULTILINE)
    label_format = re.compile(ur'\(("."(?: [0-9])+)\)+', re.MULTILINE)

    pure_lisp = None
    with open(labels_file, 'r') as l_file:
        pure_lisp = re.sub(comment_pattern, '', l_file.read().decode(encoding))
        pure_lisp = re.sub(norm_pattern, ' ', pure_lisp)
        #print_if_verbose(pure_lisp, verbose)
    labels = {l.split()[0].replace("\"", ""):sort_after_first([int(x) for x in l.split()[1:]]) for l in re.findall(label_format, pure_lisp)}
    return labels

def sort_after_first(some_list):
    #print 'sorting', some_list, some_list[0:1] + sorted(some_list[1:])
    return some_list[0:1] + sorted(some_list[1:])
    
def read_gold_data(gold_file, labels_file):
    """
    Read the gold standard based on label file.
    """
    labels = read_label_file(labels_file)
    with open(gold_file, 'r') as g_file:
        lines = [l.strip().decode(encoding) for l in g_file.readlines()]
        lines = [re.sub(comment_pattern, '', l) for l in lines]
        lines = [re.sub('[!,].+', '', l).strip() for l in lines]
        lines = [re.sub('[/-]', '+', l) for l in lines] 
        lines = filter(lambda x:len(x)>0, lines)
    if verbose:
        print_if_verbose('regex: ' + '^[' + ''.join(labels.keys()) + ' 	]+$', verbose)
            
    label_regex = re.compile('^[' + ''.join(labels.keys()) + ' 	]+$', re.MULTILINE)    #Regex to recognize label lines
    word_regex = re.compile(ur'^[0-9]+[ 	]+[^ .	]+$', re.MULTILINE)            #Regex to recognize segmentation lines
    prev_line = None   #Will hold the previous line
    gold_lines = []
    for line in lines:
        if '.' in line: #If there is a . in the line, it is an ambiguous segmentation line. 
            #print prev_line,":", line
            assert prev_line is not None #There must be a label line immediately before this line.
            if verbose:
                print_if_verbose(line + ' is ambigious with label ' + prev_line, verbose)
            gold_lines.append((prev_line, line,))
            #print prev_line, line
            #seg, hutmeg_seg = make_seg(prev_line, line, labels, consistent, label_choices, do_hutmegs)
            #segs.append(seg)
            #hutmeg_segs.append(hutmeg_seg)
            prev_line = None
    	elif re.match(label_regex, line):
            if verbose:
                print_if_verbose(line + ' is label', verbose)
            prev_line = line
    	elif re.match(word_regex, line):
            if verbose:
                print_if_verbose(line +' is non-amb', verbose)
            #seg, hutmeg_seg = make_seg(None, line, None, False, None, do_hutmegs)
            #segs.append(seg)
            #hutmeg_segs.append(hutmeg_seg)
            gold_lines.append((None,line,))
            prev_line = None
        else:
            print_error('Error |' + line + '| is unknown. Check for any labels not explained')
            raise RuntimeError()
    return gold_lines, labels


def validate_seg_and_label(seg, label, labels):
    """Validate segmentation and label"""
    assert len(label) == len(seg)-1
    if label:
        for l in label:
            assert l in labels


            
def read_and_validate_gold(gold_file, label_file):
    gold_lines, labels = read_gold_data(gold_file, label_file) #Read from file
    #Clean the segmentation lines:
    gold_lines = [(label, re.sub('^[0-9 	]+', '', seg))for label, seg in gold_lines]
    
    #split labels and segmentations
    gold_lines = [(label.split() if label else [], seg.split('.'))for label, seg in gold_lines]

    #Check length of labels
    for label, seg in gold_lines:
        validate_seg_and_label(seg, label, labels)
    #seg_line = re.sub('^[0-9 	]+', '', seg_line)
    return gold_lines, labels
    



def flen(iterable):
    return float(len(iterable))
def get_seg_set(segmentation):
    boundaries_lengths = [len(seg)  for seg in segmentation]
    boundaries_set = np.add.accumulate(boundaries_lengths)[:-1]
    return map(lambda x:x-1, boundaries_set)


def get_seg_boundaries_0_1(segmentation):
    result = [[False for s in range(len(seg)-1)] + [True] for seg in segmentation]
    result = [x for l in result for x in l][:-1]
    return result


def boundary_pra_for_word(gold_segmentation, response_segmentation):
    """Compute boundary precision, recall and accuracy for one unambiguous segmentation.
    """
    assert ''.join(gold_segmentation) == ''.join(response_segmentation)  #Make sure that the words are the same
    #Precision and recall do not care about the number of matches/mismatches in boundaries, they just look for how many matches are found.
    gold_boundaries_set = get_seg_set(gold_segmentation)
    response_boundaries_set = get_seg_set(response_segmentation)
    intersect = [boundary for boundary in gold_boundaries_set if boundary in response_boundaries_set]

    gold_boundaries_0_1 = get_seg_boundaries_0_1(gold_segmentation)
    response_boundaries_0_1 = get_seg_boundaries_0_1(response_segmentation)
    matches = [g==r for g,r in zip(response_boundaries_0_1, gold_boundaries_0_1)]
    num_matches = sum(matches)
    num_boundaries = len(gold_boundaries_0_1)
    return [len(gold_boundaries_set), len(response_boundaries_set), len(intersect), num_matches, num_boundaries]


def evaluate(gold, labels, response, verbose=True):
    if type(response[0]) is list:
        #print 'Doing it'
        response = [s[0] for s in response]
    #print response
    choices = find_preferred_choices(gold, response, labels)
    gold_reference_segs = {seg.replace('+', ''):seg.split('+') for seg in generate_segmentations_based_on_choices(gold, labels, choices)}
    response_map = {seg.replace('+', ''):seg.split('+') for seg in response if seg.replace('+', '') in gold_reference_segs}
    gold_list = [v for k,v in gold_reference_segs.iteritems()]
    response_seg = [response_map[k] for k,v in gold_reference_segs.iteritems()]
    print_if_verbose('Response', verbose)
    #print 'Response'
    print_if_verbose('\n'.join(['+'.join(x) for x in response_seg]), verbose)
    #print 'Gold'
    print_if_verbose('\n'.join(['+'.join(x) for x in gold_list]), verbose)
    stats = boundary_precision_recall_accuracy(gold_list, response_seg)
    print_if_verbose('', verbose)
    print_if_verbose(utabulate([stats], headers=['#gold bounaries', '#response boundaries', '#intersect(gold, response)', '#matches', '#possible boundaries']), verbose)
    print_if_verbose('', verbose)
    results = boundary_precision_recall_accuracy_f_score(stats)
    print_if_verbose(utabulate([results], headers=['precision', 'recall', 'f1', 'acuuracy']), verbose)
    
    return stats, results
    
    

def boundary_precision_recall_accuracy(gold_segmentations, response_segmentations):
    res = np.array(np.array([boundary_pra_for_word(gold_seg, res_seg) for gold_seg, res_seg in zip(gold_segmentations, response_segmentations)]))
    #print res
    return np.sum(res, axis=0)

def boundary_precision_recall_accuracy_f_score(statistics):
    gold_set, response_set, intersect, num_matches, num_boundaries = statistics
    precision = float(intersect)/response_set if response_set else 1.0
    recall    = float(intersect)/gold_set if gold_set else 1.0
    f_measure = 2.0 * precision * recall / (precision + recall)
    accuracy = float(num_matches)/num_boundaries
    return map(lambda x: 100*x, (precision, recall, f_measure, accuracy,))

def ilog2(n):
    n = int(n)
    assert n >= 0
    r = 0
    while n > 1:
        n /= 2
        r += 1
    return r
def align_gold_seg(gold, response, labels, label_votes):
    #print 'Aligining'
    #print gold
    #Replace labels with special characters
    str1 = '.'.join(gold[1])
    str2 = response
    #print str1, str2
    c1 = 0
    c2 = 0
    n1 = len(str1)
    n2 = len(str2)
    alignment=[]
    num_labels_read = 0
    current_label_index = -1
    chosen_value_for_current_label = 0
    while c1 < n1 and c2 < n2:
        if str1[c1] == str2[c2]:
            alignment += [(str1[c1], str2[c2],)]
            c1 += 1
            c2 += 1
        elif str1[c1] == u'.':
            current_label_index += 1
            num_labels_read += 1
            chosen_value_for_current_label = 2 * chosen_value_for_current_label
            if str2[c2] == u'+':
                alignment += [(str1[c1], str2[c2],)]
                chosen_value_for_current_label += 1
                c2 += 1
            else:
                alignment += [(str1[c1], '_',)]
            try:
                assert current_label_index <= 0 or num_labels_read == 1 or gold[0][current_label_index] == gold[0][current_label_index - 1]
            except AssertionError as ae:
                print_error(str1)
                print_error(current_label_index, gold[0][current_label_index], gold[0][current_label_index - 1], labels[gold[0][current_label_index]], num_labels_read)
                assert False
            if num_labels_read == ilog2(labels[gold[0][current_label_index]][0]):
                #reset
                num_labels_read = 0
                #print 'Register %d for %s' % (chosen_value_for_current_label, gold[0][current_label_index])
                label_votes[gold[0][current_label_index]].append(chosen_value_for_current_label)
                chosen_value_for_current_label = 0
            c1 += 1
        elif str1[c1] in u'+':
            alignment += [(str1[c1], '_',)]
            c1 += 1
        else:
            assert str2[c2] == u'+'
            alignment += [('_', str2[c2],)]
            c2 += 1
    line1 = [l1 for l1,l2 in alignment]
    line2 = [l2 for l1,l2 in alignment]
    print_if_verbose('' , verbose)
    print_if_verbose('Aligned:', verbose)
    print_if_verbose(''.join(line1), verbose)
    print_if_verbose(''.join(line2), verbose)
    return label_votes


def choose_option_by_vote(choices, votes):
    #00  10 11
    #coefficients: 2 (once), 1 (twice), 0(once) based on the chosen one
    #In case of three:
    #000 001 010 011 100 101 110 111
    #coefficients: 3 (3C3 times), 2 (3C2 times), 1 (3C1 times), 0 (3C0) times
    #for n-way:    i (nCi times), i from n to 0
    num_bits = ilog2(choices[0])
    summary = []
    for candidate in xrange(len(votes)):
        #if this candidate is chosen, what happens to other?
        contr_scores = []
        for contributor in xrange(len(votes)):
            #How many bits do they share? Use Hamming weight
            bt = ~(contributor ^ candidate) & (choices[0] - 1)
            common_bits = bin(bt).count('1')
            #print 'Common bits', candidate, contributor, common_bits, bin(bt)
            contr_scores += [common_bits*votes[candidate]]
            #cand_score += common_bits * votes[candidate]
        summary += [contr_scores]
    print_if_verbose('Votes: %s' % str(votes), verbose)
    print_if_verbose('Effective score:', verbose)
    print_if_verbose(utabulate(summary), verbose)
    summ = np.sum(summary, axis=1)
    print_if_verbose('sum ' + str(summ), verbose)
    for index, value in sorted(enumerate(summ), key=lambda x:x[1], reverse=True):
        if index in choices[1:]:
            return index
    
def find_preferred_choices(gold, response, labels):
    """Working on this!!"""
    relevant_gold = [g for g in gold]
    relevant_gold_w = {''.join(g[1]).replace('+',''):g for g in relevant_gold}
    response = {seg.replace('+', ''):seg for seg in response}
    #print response
    relevant_response = {word:seg for word,seg in response.iteritems() if word in relevant_gold_w}
    
    label_votes = defaultdict(list)
    for word in relevant_gold_w:
        #print 'Word', word
        #print 'Gold:', relevant_gold_w[word][1][0]
        #print 'Response:', relevant_response[word]
        align_gold_seg(relevant_gold_w[word], relevant_response[word], labels, label_votes)
        #print x
        #print
    label_votes = {k:Counter(v) for k,v in label_votes.iteritems()}
    label_votes_array = {}
    label_choices = {}
    #print 'Votes:',label_votes
    for label, choices in labels.iteritems():
        if label not in label_votes:
            print_if_verbose('Label {} does not appear in the gold data.'.format(label), verbose)
            label_votes[label] = {i:0 for i in range(choices[0])}
        #print label_votes[label]
        votes = [label_votes[label][i] for i in range(choices[0])]
        #print label, choices, label_votes[label], votes
        label_votes_array[label] = votes
        print_if_verbose('Label: %s(%d)'%(label,choices[0]), verbose)
        print_if_verbose('Choices: %s'%str(choices[1:]), verbose)
        chosen_index = choose_option_by_vote(choices, votes)
        label_choices[label] = chosen_index
        print_if_verbose('Chose %d for %s' % (chosen_index, label), verbose)
        print_if_verbose('', verbose)
    #print label_votes_array

    return label_choices
    
    
def generate_segmentations_based_on_choices(gold, labels, choices):
    print_if_verbose(choices, verbose)
    for word in gold:
        label, seg = word
        if not label:
            yield seg[0]
            continue
        res = seg[0]
        label_counter = 0
        current_label = label[label_counter]
        bit_number = 0# ilog2(labels[current_label][0]) - 1
        while label_counter < len(label):
            if bit_number == 0:
                current_label = label[label_counter]
                bit_number = ilog2(labels[current_label][0]) - 1
            else:
                bit_number -= 1
            bit_with_one = 1 << bit_number
            if bit_with_one & choices[current_label]:
                res += '+'
            res += seg[label_counter + 1]
            label_counter += 1
        yield res
